@@ -81,7 +81,7 @@ int main(int argc, char *argv[])
 	now = time(NULL);
 
 	/* set up initial configuration {{{ */
-	memset(&cf, NULL, sizeof(cf));
+	memset(&cf, 0, sizeof(cf));
 	cf.debug = true;
 	cf.current_display_size = 0;
 	cf.input_count = 0;
@@ -1104,11 +1104,19 @@ int dprintf(const char *fmt, ...) /* {{{ */
 {
 	FILE *d;
         va_list args;
+        static char fileName[1024] = {'\0'};
 
-	if (cf.debug && (d = fopen(DEBUG_OUTPUT, "a")))
-	{
-        	va_start(args, fmt);
-        	vfprintf(d, fmt, args);
+       	if ( !strlen( fileName ) )
+      	{
+        	strcpy( fileName, "/tmp/atop.XXXXXX" );
+        	mkdtemp( fileName );
+        	strncat( fileName, "/debug", sizeof(fileName ) );
+       	}
+
+        if (cf.debug && (d = fopen(fileName, "a")))
+        {
+                va_start(args, fmt);
+                vfprintf(d, fmt, args);
 		fclose(d);
         	va_end(args);
 	}
