@@ -8,7 +8,7 @@ extern time_t now;
 extern Circle *c;
 
 extern map *um, /* urlmap */
-// unused  *im, /* ipmap */
+           *im, /* ipmap */
            *hm, /* hostmap */
            *rm, /* referrermap */
            *fm; /* filemap */
@@ -25,6 +25,9 @@ extern map *last_display_map;
 
 bool display(time_t last_display) /* {{{ */
 {
+	struct gstat *w_gstats;
+	Circle *w_c;
+
 	if (cf.do_immed_display)
 	{
 		/* we've been asked to update immediately, whether we're
@@ -48,6 +51,9 @@ bool display(time_t last_display) /* {{{ */
 
 	display_header();
 
+	//display_histogram();
+	//return true;
+
 	if (cf.display_mode != DISPLAY_DETAIL)
 	{
 		/* normal display */
@@ -160,20 +166,18 @@ bool display(time_t last_display) /* {{{ */
 void display_header() /* {{{ */
 {
 	int itmp;
-	float bytes, ftmp;
+	float bytes, bps, per_req, ftmp;
 	unsigned int secs_offset, diff, d = 0, h = 0, m = 0, s = 0;
-	char bytes_suffix, ctmp[80];
-	struct tm *ltm;
+	char bytes_suffix, bps_suffix, per_req_suffix;
 
-	move(0,0);
+
+	move(0, 0);
 	clrtoeol();
 
 	/* last hit */
-//	ltm = localtime(&gstats.alltime.last);
-//	strftime(ctmp, sizeof(ctmp), "%H:%M:%S", ltm);
 	secs_offset = gstats.alltime.last % 86400;
 	mvprintw(0, 0, "last hit: %02d:%02d:%02d",
-	  secs_offset /3600, (secs_offset/ 60) % 60, secs_offset % 60);
+	  secs_offset / 3600, (secs_offset / 60) % 60, secs_offset % 60);
 	
 	/* uptime */
 	diff = (unsigned int)difftime(now, gstats.start);
@@ -190,25 +194,26 @@ void display_header() /* {{{ */
 	}
 
 	/* current time */
-//	ltm = localtime(&now);
-//	strftime(ctmp, sizeof(ctmp), "%H:%M:%S", ltm);
 	secs_offset = now % 86400;
 	mvprintw(0, 71, "%02d:%02d:%02d",
 	  secs_offset /3600, (secs_offset/ 60) % 60, secs_offset % 60);
 
 
 	//All: 1,140,532 requests (39.45/sec),  999,540,593 bytes (857,235/sec) 
-	bytes = readableNum(gstats.alltime.bytecount, &bytes_suffix);
 	ftmp = getMAX(now-gstats.alltime.first, 1); /* divide-by-zero hack */
+	bytes = readableNum(gstats.alltime.bytecount, &bytes_suffix);
+	bps = readableNum(gstats.alltime.bytecount/ftmp, &bps_suffix);
+	per_req = readableNum(
+	    gstats.alltime.bytecount/getMAX(gstats.alltime.reqcount, 1),
+	    &per_req_suffix);
 	attron(A_BOLD);
 	mvprintw(1, 0, 
-	  "All: %12.0f reqs (%6.1f/sec)  %11.1f%c (%6.1fK/sec)  (%7.1fK/req)",
+	  "All: %12.0f reqs (%6.1f/sec)  %11.1f%c (%7.1f%c/sec)   %7.1f%c/req",
 	  gstats.alltime.reqcount,
 	  gstats.alltime.reqcount/ftmp,
 	  bytes, bytes_suffix,
-	  (gstats.alltime.bytecount/1024)/ftmp,
-	  (float)(gstats.alltime.bytecount/1024)/
-	      getMAX(gstats.alltime.reqcount, 1));
+	  bps, bps_suffix,
+	  per_req, per_req_suffix);
 	attroff(A_BOLD);
 
 
@@ -218,12 +223,26 @@ void display_header() /* {{{ */
 	       gstats.r_codes[4].reqcount + gstats.r_codes[5].reqcount;
 	if (ftmp == 0) ftmp = 1; /* avoid NaN with no hits */
 	mvprintw(2, 0,
-	  "2xx: %7.0f (%4.1f%%) 3xx: %7.0f (%4.1f%%) "
-	  "4xx: %5.0f (%4.1f%%) 5xx: %5.0f (%4.1f%%) ",
-	  gstats.r_codes[2].reqcount, (gstats.r_codes[2].reqcount/ftmp)*100,
-	  gstats.r_codes[3].reqcount, (gstats.r_codes[3].reqcount/ftmp)*100,
-	  gstats.r_codes[4].reqcount, (gstats.r_codes[4].reqcount/ftmp)*100,
-	  gstats.r_codes[5].reqcount, (gstats.r_codes[5].reqcount/ftmp)*100);
+	 "2xx: %7.0f (%4.*f%%) 3xx: %7.0f (%4.*f%%) "
+	 "4xx: %5.0f (%4.*f%%) 5xx: %5.0f (%4.*f%%) ",
+
+	 gstats.r_codes[2].reqcount,
+	 (gstats.r_codes[2].reqcount/ftmp) == 1 ? 0 : 1,
+	 (gstats.r_codes[2].reqcount/ftmp)*100,
+
+	 gstats.r_codes[3].reqcount,
+	 (gstats.r_codes[3].reqcount/ftmp) == 1 ? 0 : 1,
+	 (gstats.r_codes[3].reqcount/ftmp)*100,
+	 
+	 gstats.r_codes[4].reqcount,
+	 (gstats.r_codes[4].reqcount/ftmp) == 1 ? 0 : 1,
+	 (gstats.r_codes[4].reqcount/ftmp)*100,
+
+	 gstats.r_codes[5].reqcount,
+	 (gstats.r_codes[5].reqcount/ftmp) == 1 ? 0 : 1,
+	 (gstats.r_codes[5].reqcount/ftmp)*100
+	 
+	);
 
 	/* housecleaning on the circle, if its required in this class */
 	c->updatestats();
@@ -231,28 +250,45 @@ void display_header() /* {{{ */
 	itmp = now - c->oldest();
 	itmp = getMAX(itmp, 1); /* divide-by-zero hack */
 	bytes = readableNum(c->getbytecount(), &bytes_suffix);
+	bps = readableNum(c->getbytecount()/itmp, &bps_suffix);
+	per_req = readableNum(c->getbytecount()/getMAX(c->getreqcount(), 1),
+	    &per_req_suffix);
 	attron(A_BOLD);
 	mvprintw(3, 0,
-	  "R (%3ds): %7.0f reqs (%6.1f/sec)  %11.1f%c (%6.1fK/sec)  (%7.1fK/req)",
+	  "R (%3ds): %7.0f reqs (%6.1f/sec)  %11.1f%c (%7.1f%c/sec)   %7.1f%c/req",
 	  itmp, c->getreqcount(),
 	  ((float)c->getreqcount()/itmp),
 	  bytes, bytes_suffix,
-	  ((float)(c->getbytecount()/1024)/itmp),
-	  (float)(c->getbytecount()/1024)/getMAX(c->getreqcount(), 1)
-	  );
+	  bps, bps_suffix,
+	  per_req, per_req_suffix
+	);
 	attroff(A_BOLD);
 
 	ftmp = c->getsummary(2) + c->getsummary(3) +
 	       c->getsummary(4) + c->getsummary(5);
 	if (ftmp == 0) ftmp = 1; /* avoid NaN with no hits */
 	mvprintw(4, 0,
-	  "2xx: %7.0f (%4.1f%%) 3xx: %7.0f (%4.1f%%) "
-	  "4xx: %5.0f (%4.1f%%) 5xx: %5.0f (%4.1f%%) ",
-	  c->getsummary(2), (c->getsummary(2)/ftmp)*100,
-	  c->getsummary(3), (c->getsummary(3)/ftmp)*100,
-	  c->getsummary(4), (c->getsummary(4)/ftmp)*100,
-	  c->getsummary(5), (c->getsummary(5)/ftmp)*100);
+	  "2xx: %7.0f (%4.*f%%) 3xx: %7.0f (%4.*f%%) "
+	  "4xx: %5.0f (%4.*f%%) 5xx: %5.0f (%4.*f%%) ",
+	 c->getsummary(2),
+	 (c->getsummary(2)/ftmp) == 1 ? 0 : 1,
+	 (c->getsummary(2)/ftmp)*100,
 
+	 c->getsummary(3),
+	 (c->getsummary(3)/ftmp) == 1 ? 0 : 1,
+	 (c->getsummary(3)/ftmp)*100,
+
+	 c->getsummary(4),
+	 (c->getsummary(4)/ftmp) == 1 ? 0 : 1,
+	 (c->getsummary(4)/ftmp)*100,
+
+	 c->getsummary(5),
+	 (c->getsummary(5)/ftmp) == 1 ? 0 : 1,
+	 (c->getsummary(5)/ftmp)*100
+	);
+
+//	mvprintw(5, 0,
+//	    "Unique Objects:         Size Footprint:");
 
 	/* if any filters are active, and the user is not in a submenu,
 	 * display a summary */
@@ -273,7 +309,8 @@ void display_header() /* {{{ */
 		if (*active_filters)
 		{
 			y = COLS - (strlen(active_filters) + 11);
-			mvprintw(5, y, "Filtering: %s", active_filters);
+			mvprintw(SUBMENU_LINE_NUMBER,
+			         y, "Filtering: %s", active_filters);
 		}
 
 	}
@@ -282,7 +319,7 @@ void display_header() /* {{{ */
 
 void display_list() /* {{{ */
 {
-	int x;
+	int x, t, xx;
 	char *cptmp;
 	struct config scf; /* for copying cf, see comment below at memcpy */
 
@@ -292,20 +329,23 @@ void display_list() /* {{{ */
 	struct itemlist *item_ptr = NULL;
 
 	/* for easy reference during walk() */
-	map *map; unsigned int pos, hash;
+	map *map; unsigned int hash;
+	int pos;
 
 	/* walk() pointer */
 	struct logbits *lb;
 
-	if (c->getreqcount() == 0)
+	/* make an array containing all we have, then sort it */
+	items_size = c->getreqcount();
+	if (items_size == 0)
 	{
 		/* nothing to do! */
+		move(LINES_RESERVED-1, 0);
+		clrtobot();
 		refresh();
 		return;
 	}
 
-	/* make an array containing all we have, then sort it */
-	items_size = c->getreqcount();
 	item_hash.create( (int)items_size * 5 );
 
 	/* this is our array; the cast is because items_size is a double but
@@ -314,7 +354,7 @@ void display_list() /* {{{ */
 	if (items) free(items); /* get rid of the last one */
 	items = (struct itemlist *)
 	    calloc((unsigned int)items_size, sizeof(itemlist));
-	
+
 	/* another thread may change the contents of cf while we're running,
 	 * and it would be undesirable to have most of this change on us, so
 	 * we make safe copies to use */
@@ -353,13 +393,25 @@ void display_list() /* {{{ */
 		switch(scf.display_mode)
 		{
 			default:
-			case DISPLAY_URLS: hash = lb->url_hash; pos = lb->url_pos; break;
-			case DISPLAY_HOSTS: hash = lb->host_hash; pos = lb->host_pos; break;
-			case DISPLAY_REFS: hash = lb->ref_hash; pos = lb->ref_pos; break;
-			//case DISPLAY_FILES: hash = lb->file_hash ; pos = lb->file_pos; break;
+			case DISPLAY_URLS:
+				hash = lb->url_hash; pos = lb->url_pos;
+				break;
+
+			case DISPLAY_HOSTS:
+				hash = lb->host_hash; pos = lb->host_pos;
+				break;
+
+			case DISPLAY_REFS:
+				hash = lb->ref_hash; pos = lb->ref_pos;
+				break;
+#if HAVE_FILE_MODE_DISPLAY
+			case DISPLAY_FILES:
+				hash = lb->file_hash ; pos = lb->file_pos;
+				break;
+#endif
 		}
 
-		/* FILTERS */
+		/* FILTERS {{{ */
 		/* skip this item if it doesn't match our filter */
 		if (cf.urlfilter->isactive() &&
 		    !cf.urlfilter->match(um->reverse(lb->url_pos)))
@@ -372,7 +424,7 @@ void display_list() /* {{{ */
 		if (cf.reffilter->isactive() &&
 		    !cf.reffilter->match(rm->reverse(lb->ref_pos)))
 			continue;
-
+		/* }}} */
 
 		/* look up whatever string this pos is for */
 		cptmp = map->reverse(pos);
@@ -380,15 +432,21 @@ void display_list() /* {{{ */
 		/* then lookup that string in items */
 		item_ptr = (struct itemlist *)item_hash.lookup(cptmp);
 		
+
 		/* do we have this string already? */
 		if (item_ptr == NULL)
 		{
 			/* not seen it, make a new slot */
 			item_ptr = &items[item_used];
 
-			item_hash.insert(map->reverse(pos), &items[item_used]);
+			item_hash.insert(cptmp, item_ptr);
 			item_ptr->item = pos;
 
+			/* store the ip position in the itemlist too, just
+			** in case we have to look it up in show_map_line
+			*/
+			item_ptr->ip_item = lb->ip_pos;
+
 			item_ptr->hash = hash;
 
 			++item_used;
@@ -397,9 +455,15 @@ void display_list() /* {{{ */
 		}
 
 		/* we have the string in items now, so update stats in array */
-		++(item_ptr->reqcount);
+		item_ptr->reqcount++;
 		item_ptr->bytecount += lb->bytes;
 
+		/* we wish to count up how many times a given retcode
+		 * has occurred for this item */
+		t = (int)(lb->retcode/100);
+		item_ptr->r_codes[t].reqcount++;
+		item_ptr->r_codes[t].bytecount += lb->bytes;
+
 		if (lb->time < item_ptr->first)
 			item_ptr->first = lb->time;
 	}
@@ -416,15 +480,37 @@ void display_list() /* {{{ */
 		if (item_ptr->timespan == 0)
 			item_ptr->timespan = 1; /* hack to avoid /0 */
 
+		for(xx = 2 ; xx < 6 ; xx++)
+		{
+			item_ptr->r_codes[xx].rps = ((float)
+			    item_ptr->r_codes[xx].reqcount/item_ptr->timespan);
+
+			item_ptr->r_codes[xx].bps = ((float)
+			    item_ptr->r_codes[xx].bytecount/item_ptr->timespan);
+		}
+
 		item_ptr->kbps = ((float)
 		    (item_ptr->bytecount/1024)/item_ptr->timespan);
 		item_ptr->rps = ((float)
 		    item_ptr->reqcount/item_ptr->timespan);
 	}
 
-	move(6, 0);
+	move(SUBMENU_LINE_NUMBER+1, 0);
 	clrtobot();
 
+	/* a suitable header, at LINES_RESERVED-1
+	 * (-1 because curses starts at zero, but I've started at 1) */
+	switch(scf.numbers_mode)
+	{
+		case NUMBERS_HITS_BYTES:
+		mvaddstr(LINES_RESERVED-1, 0, " REQS REQ/S    KB KB/S");
+		break;
+
+		case NUMBERS_RETCODES:	
+		mvaddstr(LINES_RESERVED-1, 0, "  2xx   3xx   4xx  5xx");
+		break;
+	}
+
 	/* what are we showing in the table? we have a few options;
 	 *
 	 * top $X URLS or IPs/Hosts or Referrers
@@ -433,17 +519,17 @@ void display_list() /* {{{ */
 	if (cf.display_mode == DISPLAY_DETAIL)
 	{
 		/* detailed referrer stats for a given URL */
-		mvaddstr(LINES_RESERVED-1, 0, "REQS REQ/S    KB  KB/S");
+	//	mvaddstr(LINES_RESERVED-1, 0, "REQS REQ/S    KB  KB/S");
 
 		/* display only the item we're interested in */
 		for (x = 0 ; x < item_used ; ++x)
 		{
-			dprintf("%d %d\n", item_ptr->hash, cf.selected_item_hash);
 			item_ptr = &items[x];
 			if (item_ptr->hash == cf.selected_item_hash)
 			{
 				/* show stats for this line only */
-				show_map_line(item_ptr, 0, map, NO_INDENT);
+				show_map_line(item_ptr, 0, map,
+				    NO_INDENT, scf.numbers_mode);
 				break;
 			}
 		}
@@ -451,26 +537,23 @@ void display_list() /* {{{ */
 		return;
 	}
 
-
 	/* top $X items; sort the array for display */
-	if (item_used) shellsort(items, item_used, scf.sort);
+	if (item_used) shellsort_wrapper(items, item_used, scf);
 
 	/* display something */
 
-	/* a suitable header, at LINES_RESERVED-1
-	 * (-1 because curses starts at zero, but I've started at 1) */
 	switch(scf.display_mode)
 	{
 		case DISPLAY_URLS:
-		mvaddstr(LINES_RESERVED-1, 0, "REQS REQ/S    KB  KB/S URL");
+		mvaddstr(LINES_RESERVED-1, 23, "URL");
 		break;
 	
 		case DISPLAY_HOSTS:
-		mvaddstr(LINES_RESERVED-1, 0, "REQS REQ/S    KB  KB/S HOST");
+		mvaddstr(LINES_RESERVED-1, 23, "HOST");
 		break;
 
 		case DISPLAY_REFS:
-		mvaddstr(LINES_RESERVED-1, 0, "REQS REQ/S    KB  KB/S REFERRER");
+		mvaddstr(LINES_RESERVED-1, 23, "REFERRER");
 		break;
 	}
 
@@ -485,7 +568,8 @@ void display_list() /* {{{ */
 
 		/* render the line itself at position disp. map should be
 		 * already set from earlier in this function */
-		show_map_line(item_ptr, disp, map, NO_INDENT);
+		show_map_line(item_ptr, disp, map,
+		    NO_INDENT, scf.numbers_mode);
 
 		++disp;
 	}
@@ -522,7 +606,10 @@ void display_sub_list(short display_mode_override, /* {{{ */
 	 * URL or IP or REFERRER.
 	*/
 
-	int x;
+	/* I'd like to lose this code and replace the breakdown screen
+	 * with something more useful.. */
+
+	int x, t, xx;
 	char *cptmp;
 	unsigned int h;
 	struct config scf; /* for copying cf, see comment below at memcpy */
@@ -618,23 +705,35 @@ void display_sub_list(short display_mode_override, /* {{{ */
 			/* not seen it, make a new slot */
 			item_ptr = &subitems[item_used];
 
-			item_hash.insert(map->reverse(pos), &subitems[item_used]);
+			item_hash.insert(cptmp, item_ptr);
 			item_ptr->item = pos;
 
-			++item_used;
+			/* store the ip position in the itemlist too, just
+			** in case we have to look it up in show_map_line
+			*/
+			item_ptr->ip_item = lb->ip_pos;
+
 			item_ptr->last = now;
 			item_ptr->first = lb->time;
+
+			++item_used;
 		}
 
 		/* we have the string in items now, so update stats in array */
 		++(item_ptr->reqcount);
 		item_ptr->bytecount += lb->bytes;
 
+		/* we wish to count up how many times a given retcode
+		 * has occurred for this item */
+		t = (int)(lb->retcode/100);
+		item_ptr->r_codes[t].reqcount++;
+		item_ptr->r_codes[t].bytecount += lb->bytes;
+
 		if (lb->time < item_ptr->first)
 			item_ptr->first = lb->time;
 	}
 
-	/* no further use for this hash */
+	/* no further use for this lookup hash */
 	item_hash.destroy();
 
 	/* calculate some stats; timespan, kbps, rps */
@@ -646,6 +745,15 @@ void display_sub_list(short display_mode_override, /* {{{ */
 		if (item_ptr->timespan == 0)
 			item_ptr->timespan = 1; /* hack to avoid /0 */
 
+		for(xx = 2 ; xx < 6 ; xx++)
+		{
+			item_ptr->r_codes[xx].rps = ((float)
+			    item_ptr->r_codes[xx].reqcount/item_ptr->timespan);
+
+			item_ptr->r_codes[xx].bps = ((float)
+			    item_ptr->r_codes[xx].bytecount/item_ptr->timespan);
+		}
+
 		item_ptr->kbps = ((float)
 		    (item_ptr->bytecount/1024)/item_ptr->timespan);
 		item_ptr->rps = ((float)
@@ -653,7 +761,8 @@ void display_sub_list(short display_mode_override, /* {{{ */
 	}
 
 	/* top $X items; sort the array for display */
-	if (item_used) shellsort(subitems, item_used, scf.sort);
+	if (item_used) shellsort_wrapper(subitems, item_used, scf);
+
 
 	//clrtobot();
 
@@ -664,16 +773,16 @@ void display_sub_list(short display_mode_override, /* {{{ */
 	switch(display_mode_override)
 	{
 		case DISPLAY_URLS:
-		mvaddstr(LINES_RESERVED+offset-1, 23, "URL");
-		break;
+			mvaddstr(LINES_RESERVED+offset-1, 23, "URL");
+			break;
 	
 		case DISPLAY_HOSTS:
-		mvaddstr(LINES_RESERVED+offset-1, 23, "HOST");
-		break;
+			mvaddstr(LINES_RESERVED+offset-1, 23, "HOST");
+			break;
 
 		case DISPLAY_REFS:
-		mvaddstr(LINES_RESERVED+offset-1, 23, "REFERRER");
-		break;
+			mvaddstr(LINES_RESERVED+offset-1, 23, "REFERRER");
+			break;
 	}
 
 	disp = offset; /* count how many we've shown */
@@ -686,7 +795,7 @@ void display_sub_list(short display_mode_override, /* {{{ */
 
 		/* render the line itself at position disp. map should be
 		 * already set from earlier in this function */
-		show_map_line(item_ptr, disp, map, 2);
+		show_map_line(item_ptr, disp, map, 2, scf.numbers_mode);
 
 		++disp;
 	}
@@ -700,21 +809,16 @@ void translate_screen_to_pos() /* {{{ */
 
 	/* convert cf.selected_item_screen to cf.selected_item_pos */
 	cf.selected_item_pos = items[cf.selected_item_screen].item;
-	if (cf.selected_item_pos)
-	{
+
+//	/* cf.selected_item_pos may well be zero */
+//	if (cf.selected_item_pos)
+//	{
 		/* make a hash of it */
 		cf.selected_item_hash =
 	 	   TTHash(last_display_map->reverse(cf.selected_item_pos));
 
 		cf.selected_item_mode = cf.display_mode;
-
-#if 0
-		move(5,0);
-		clrtoeol();
-		mvprintw(5,0, "%s",
-		    last_display_map->reverse(cf.selected_item_pos));
-#endif
-	}
+//	}
 } /* }}} */
 
 void drawMarker(void) /* update position of asterisk next to URLs {{{ */
@@ -736,7 +840,7 @@ void drawMarker(void) /* update position of asterisk next to URLs {{{ */
 	mvaddch(LINES_RESERVED+cf.selected_item_screen+1, COLS_RESERVED-3, ' ');
 
 	/* come to rest under header */
-	move(5, 0);
+	move(SUBMENU_LINE_NUMBER, 0);
 	
 	refresh();
 } /* }}} */
@@ -745,35 +849,114 @@ void drawMarker(void) /* update position of asterisk next to URLs {{{ */
 /* render contents of item_ptr into a statistics line onscreen,
  * at offset vert_location; pull the text portion out of m */
 void show_map_line(struct itemlist *item_ptr, int vert_location, /* {{{ */
-    map *m, unsigned short indent)
+    map *m, unsigned short indent, unsigned short number_mode)
 {
-	char spacer[10] = "         "; /* nicer way to do this? */
-	spacer[indent] = (char) NULL;
+	if (number_mode == NUMBERS_HITS_BYTES)
+	{
+		/* start drawing at LINES_RESERVED */
+		mvprintw(LINES_RESERVED+vert_location, 0,
+		    "%5.0f %5.*f %5.*f %4.*f",
+		    item_ptr->reqcount,
 
-	/* start drawing at LINES_RESERVED */
-	mvprintw(LINES_RESERVED+vert_location, 0,
-	    "%4.0f %5.*f %5.*f %5.*f %s%.*s",
-	    item_ptr->reqcount,
+		    /* set 1dp if rps > 99, or 2 if not */
+		    ((item_ptr->rps > 99) ? 1 : 2),
+		    item_ptr->rps,
 
-	    /* set 1dp if rps > 99, or 2 if not */
-	    ((item_ptr->rps > 99) ? 1 : 2),
-	    item_ptr->rps,
+		    /* scale KB display; if >1000K lose the decimal point */
+		    ((item_ptr->bytecount > (999*1024)) ? 0 : 1),
+		    ((float)item_ptr->bytecount/1024),
+		    
+		    /* scale KB/s display; if >1000K/s lose the decimal point */
+		    ((item_ptr->kbps > 99) ? 0 : 1),
+		    item_ptr->kbps);
+	}
+	else /* number_mode == NUMBERS_RETCODES */
+	{
+		mvprintw(LINES_RESERVED+vert_location, 0,
+		    "%5.0f %5.0f %5.0f %4.0f",
+		    item_ptr->r_codes[2].reqcount,
+		    item_ptr->r_codes[3].reqcount,
+		    item_ptr->r_codes[4].reqcount,
+		    item_ptr->r_codes[5].reqcount);
+	}
 
-	    /* scale KB display; if >1000K lose the decimal point */
-	    ((item_ptr->bytecount > (999*1024)) ? 0 : 1),
-	    ((float)item_ptr->bytecount/1024),
-	    
-	    /* scale KB/s display; if >1000K/s lose the decimal point */
-	    ((item_ptr->kbps > 999) ? 0 : 1),
-	    item_ptr->kbps,
+	/* text after the number columns */
+
+	/* if we are displaying host/ip rather than anything else, we
+	** have slightly more work to do; we need to generate a line of the
+	** format host* [ip]
+	** * only needs to be present if it's still resolving.
+	** ip only needs to be present if list_show_ip is true.
+	*/
+
+	/* 23+indent = start at 23, but move to the right by indent
+	** spaces if we are doing, for example, a sublist
+	*/
+	if (m == hm)
+	{
+		/* we are displaying a host line */
+		char *h = NULL, *i = NULL;
+		int host_item, ip_item;
+
+		/* 200 chars ought to be enough? */
+		#define MAX_IP_STR_WIDTH 200
+
+		char str[MAX_IP_STR_WIDTH];
+
+		host_item = item_ptr->item;
+		ip_item = item_ptr->ip_item;
+
+		if (host_item >= 0) h = m->reverse(host_item);
+		if (ip_item >= 0) i = im->reverse(ip_item);
+
+		if (h && i)
+			snprintf(str, MAX_IP_STR_WIDTH, "%s [%s]", h, i);
+		else if (h)
+			snprintf(str, MAX_IP_STR_WIDTH, "%s", h);
+		else if (i)
+			snprintf(str, MAX_IP_STR_WIDTH, "[%s]", i);
+		else
+			return; /* shouldn't get reached */
+
+		mvprintw(LINES_RESERVED+vert_location, 23+indent, "%.*s",
+		    (COLS-COLS_RESERVED /* width */), str);
+	}
+	else
+	{
+		char *str = m->reverse(item_ptr->item);
+		mvprintw(LINES_RESERVED+vert_location, 23+indent, "%.*s",
+		    (COLS-COLS_RESERVED /* width */), str);
+	}
 
-	    spacer,
 
-	    (COLS-COLS_RESERVED /* width */), m->reverse(item_ptr->item));
 } /* }}} */
 
+void shellsort_wrapper(struct itemlist *items, unsigned int size, /* {{{ */
+    struct config pcf)
+{
+	short sort_method;
+
+	if (size == 0) return;
+
+	/* what are we displaying in the numbers column? */
+	switch(pcf.numbers_mode)
+	{
+		/* sorting by hits or bytes */
+		case NUMBERS_HITS_BYTES:
+			sort_method = pcf.sort;
+			break;
+		
+		/* sorting by a return code */
+		case NUMBERS_RETCODES:
+			sort_method = pcf.retcodes_sort;
+			break;
+	}
+
+	shellsort(items, size, sort_method);
+} /* }}} */
 void shellsort(struct itemlist *items, unsigned int size, int sorttype) /* {{{ */
 {
+	int x;
 	bool done;
 	unsigned int i, j, jmp = size;
 	double i_c, j_c;
@@ -789,25 +972,43 @@ void shellsort(struct itemlist *items, unsigned int size, int sorttype) /* {{{ *
 			for (j = 0; j < (size-jmp); j++)
 			{
 				i = j + jmp;
-				switch(sorttype)
+				if (cf.numbers_mode == NUMBERS_HITS_BYTES)
 				{
-					default:
-					case SORT_REQCOUNT:
+					/* numbers_mode = NUMBERS_HITS_BYTES */
+					switch(sorttype)
+					{
+						default:
+						case SORT_REQCOUNT:
 						i_c = items[i].reqcount;
 						j_c = items[j].reqcount;
 						break;
-					case SORT_REQPERSEC:
+
+						case SORT_REQPERSEC:
 						i_c = items[i].rps;
 						j_c = items[j].rps;
 						break;
-					case SORT_BYTECOUNT:
+
+						case SORT_BYTECOUNT:
 						i_c = items[i].bytecount;
 						j_c = items[j].bytecount;
 						break;
-					case SORT_BYTESPERSEC:
+
+						case SORT_BYTESPERSEC:
 						i_c = items[i].kbps;
 						j_c = items[j].kbps;
 						break;
+					}
+				}
+				else
+				{
+					/* numbers_mode = NUMBERS_RETCODES */
+					/* this is easier ;)
+					 * see apachetop.h for explanation
+					 * of what SORTTYPE_OFFSET_HACK
+					 * is */
+					x = sorttype-SORTTYPE_OFFSET_HACK;
+					i_c = items[i].r_codes[x].reqcount;
+					j_c = items[j].r_codes[x].reqcount;
 				}
 				
 				if (i_c > j_c)
@@ -826,33 +1027,44 @@ void shellsort(struct itemlist *items, unsigned int size, int sorttype) /* {{{ *
 
 float readableNum(double num, char *suffix) /* {{{ */
 {
-	#define MAX_IN_KB (1024*1024)*10
+	#define AP_TEN_KB ((double)(1024)*10)
+	#define AP_TEN_MB ((double)(1024*1024)*10)
+	#define AP_TEN_GB ((double)(1024*1024*1024)*10)
 
-	if (num < MAX_IN_KB)
+	if (num > AP_TEN_GB)
+	{
+		*suffix = 'G';
+		return (float)num/((double)(1024*1024*1024));
+	}
+	if (num > AP_TEN_MB)
+	{
+		*suffix = 'M';
+		return (float)num/((double)(1024*1024));
+	}
+	if (num > AP_TEN_KB)
 	{
 		*suffix = 'K';
 		return (float)num/1024;
 	}
-
-	*suffix = 'M';
-	return (float)num/(1024*1024);
+	
+	*suffix = 'B';
+	return (float)num;
 } /* }}} */
 
 void display_submenu_banner(char *title, int title_len, char *banner)
 {
-
 	attron(A_REVERSE);
-	mvaddstr(5, 1, title);
+	mvaddstr(SUBMENU_LINE_NUMBER, 1, title);
 	attroff(A_REVERSE);
-	mvaddstr(5, 1+title_len, banner);
-	move(5,0);
+	mvaddstr(SUBMENU_LINE_NUMBER, 1+title_len, banner);
+	move(SUBMENU_LINE_NUMBER, 0);
 	refresh();
 }
 
 void clear_submenu_banner(void)
 {
 	/* remove submenu banner */
-	move(5,0);
+	move(SUBMENU_LINE_NUMBER, 0);
 	clrtoeol();
 	refresh();
 }
@@ -867,7 +1079,8 @@ void display_help(void)
 	
 	move(2, 0);
 	addstr("ONE-TOUCH COMMANDS\n");
-	addstr("d          : switch display between urls/referrers/hosts\n");
+	addstr("d          : switch item display between urls/referrers/hosts\n");
+	addstr("n          : switch numbers display between hits & bytes or return codes\n");
 	addstr("h or ?     : this help window\n");
 	addstr("p          : (un)pause display (freeze updates)\n");
 	addstr("q          : quit ApacheTop\n");
@@ -875,13 +1088,14 @@ void display_help(void)
 	addstr("right/left : enter/exit detailed subdisplay mode\n");
 	addstr("\n");
 	addstr("SUBMENUS:\n");
-	addstr("s > SORT BY:\n");
-	addstr("\tr) requests  R) reqs/sec  b) bytes  B) byes/sec\n\n");
-	addstr("t > TOGGLE SUBDISPLAYS ON/OFF:\n");
+	addstr("s:  SORT BY: [the appropriate menu will appear for your display]\n");
+	addstr("\tr) requests  R) reqs/sec  b) bytes  B) bytes/sec\n");
+	addstr("\t2) 2xx   3) 3xx   4) 4xx   5) 5xx\n\n");
+	addstr("t:  TOGGLE SUBDISPLAYS ON/OFF:\n");
 	addstr("\tu) urls  r) referrers  h) hosts\n\n");
-	addstr("f > MANIPULATE FILTERS:\n");
-	addstr("\ta) add/edit menu c) clear all  s) show active\n");
-	addstr("\ta > ADD FILTER SUBMENU\n");
+	addstr("f:  MANIPULATE FILTERS:\n");
+	addstr("\ta) add/edit menu c) clear all  s) show active (not done yet)\n");
+	addstr("\ta:  ADD FILTER SUBMENU\n");
 	addstr("\t\tu) to urls  r) to referrers  h) to hosts\n");
 	addstr("\n");
 	attron(A_REVERSE);
@@ -910,3 +1124,92 @@ void display_active_filters()
 	cf.display_paused = true;
 
 }
+
+void display_histogram()
+{
+	int hist_height, hist_width;
+	int i, j, age, oldest;
+	//int y_scale;
+	float y_scale, y_decr;
+	float x_scale, max_bar = 0;
+	char horiz_line[128];
+
+	struct logbits *lb;
+
+	/* histogram starts at LINES_RESERVED+10 */
+	#define HISTOGRAM_START LINES_RESERVED+3
+	hist_height = 10;
+	hist_width = 60;
+
+	/* for every width character, figure out how many
+	 * characters high to draw the barchart
+	*/
+	float bar_height[hist_width];
+	char line[hist_width];
+	for(i = 0 ; i < hist_width ; i++)
+		bar_height[i] = 0;
+
+	/* figure out scales; includes divide-by-zero avoidance hack */
+	if (cf.circle_mode == TIMED_CIRCLE)
+		/* timed_circle = we know exactly how old we're going to get */
+		x_scale = ((float)hist_width/cf.circle_size);
+	else
+		x_scale = ((float)hist_width/getMAX(now - c->oldest(), 1));
+
+	/* don't scale when we have less data than we have room for */
+	if (x_scale > 1) x_scale = 1;
+
+	while(c->walk(&lb) != -1)
+	{
+		if (!lb)
+			continue;
+
+		/* we have hist_width bars, and we need to put the entire
+		 * circle into that many bars. Devise which bar we're using
+		 * for this particular lb->time */
+		age = int( x_scale * (now - lb->time) );
+
+		/* add on x_scale; this is because if we are displaying 2
+		 * seconds worth of data in one line, we only want to add on
+		 * half. */
+		bar_height[age] += x_scale;
+	}
+
+	/* find the maximum bar height we have. */
+	for(i = 0 ; i < hist_width ; ++i)
+		max_bar = getMAX(max_bar, bar_height[i]);
+
+	y_scale = max_bar;
+	y_decr = ((float)y_scale / hist_height);
+
+	for(i = 0 ; i < hist_height ; ++i)
+	{
+		if (i % 2 == 0)
+			mvprintw(HISTOGRAM_START + i, 0, "%3.0f", y_scale);
+
+		mvaddch(HISTOGRAM_START + i, 3, '|');
+
+		/* compose a row of hashes */
+		memset(line, ' ', hist_width);
+		line[hist_width] = (char)NULL;
+		for(j = 0 ; j < hist_width ; ++j)
+		{
+			if (bar_height[j] > y_scale)
+				line[j] = '#';
+		}
+		mvprintw(HISTOGRAM_START + i, 4, "%s", line);
+
+		y_scale -= y_decr;
+	}
+
+	memset(horiz_line, '-', hist_width);
+	horiz_line[hist_width] = (char)NULL;
+	mvprintw(HISTOGRAM_START + hist_height, 2, "0+%*s",
+	    hist_width, horiz_line);
+	
+	mvprintw(HISTOGRAM_START + hist_height+1, 4, "NOW");
+	mvprintw(HISTOGRAM_START + hist_height+1, hist_width+3, "-%ds",
+	    now - c->oldest());
+
+	refresh();
+}
