@@ -9,14 +9,18 @@
                          mvaddstr(x, y, "paused"); \
                          attroff(A_REVERSE)
 
+#define SUBMENU_LINE_NUMBER LINES_RESERVED-2
+
 /* display() makes an array of these, sorts, and displays */
 struct itemlist {
 	unsigned int hash;
-	unsigned int item;
+	int item, ip_item;
 	double reqcount;
 	double bytecount;
 	time_t first, last, timespan;
 	float rps, kbps;
+
+	struct hitinfo r_codes[6];
 };
 
 bool display(time_t last_display);
@@ -33,9 +37,12 @@ void drawMarker(void);
 
 #define NO_INDENT 0
 void show_map_line(struct itemlist *item_ptr, int vert_location, map *m,
-    unsigned short indent);
+    unsigned short indent, unsigned short number_mode);
 
+void shellsort_wrapper(struct itemlist *items, unsigned int size,
+    struct config pcf);
 void shellsort(struct itemlist *items, unsigned int size, int sorttype);
+
 float readableNum(double num, char *suffix);
 
 void display_submenu_banner(char *title, int title_len, char *banner);
@@ -44,4 +51,6 @@ void clear_submenu_banner(void);
 void display_help(void);
 void display_active_filters(void);
 
+void display_histogram();
+
 #endif
